@@ -46,7 +46,7 @@
 #endif
 
 static char const server_logstring[] = "tls";
-static char const client_logstring[] = "tls-client";
+static char const client_logstring[] = "tls_client";
 
 static int client_verify_certificate(gnutls_session_t tls_session);
 
@@ -171,7 +171,7 @@ void tls_close(struct tls_common_ctx *ctx)
 
 	if (ctx->handshake_state == TLS_HS_DONE) {
 		kr_log_verbose("[%s] closing tls connection to `%s`\n",
-			       ctx->client_side ? "tls-client" : "tls",
+			       ctx->client_side ? "tls_client" : "tls",
 			       kr_straddr(&ctx->session->peer.ip));
 		ctx->handshake_state = TLS_HS_CLOSING;
 		gnutls_bye(ctx->tls_session, GNUTLS_SHUT_RDWR);
@@ -280,15 +280,15 @@ int tls_process(struct worker_ctx *worker, uv_stream_t *handle, const uint8_t *b
 		int err = gnutls_handshake(tls_p->tls_session);
 		if (err == GNUTLS_E_SUCCESS) {
 			tls_p->handshake_state = TLS_HS_DONE;
+			kr_log_verbose("[%s] TLS handshake with %s has completed\n",
+				       logstring,  kr_straddr(&session->peer.ip));
 			if (tls_p->handshake_cb) {
 				tls_p->handshake_cb(tls_p->session, 0);
 			}
-			kr_log_verbose("[%s] TLS handshake with %s has completed\n",
-				       logstring,  kr_straddr(&session->peer.ip));
 		} else if (err == GNUTLS_E_AGAIN) {
 			return 0;
 		} else if (gnutls_error_is_fatal(err)) {
-			kr_log_error("[%s] gnutls_handshake failed: %s (%d)\n",
+			kr_log_verbose("[%s] gnutls_handshake failed: %s (%d)\n",
 				     logstring,
 			             gnutls_strerror_name(err), err);
 			if (tls_p->handshake_cb) {
@@ -306,7 +306,7 @@ int tls_process(struct worker_ctx *worker, uv_stream_t *handle, const uint8_t *b
 		} else if (count == GNUTLS_E_INTERRUPTED) {
 			continue; /* Try reading again */
 		} else if (count < 0) {
-			kr_log_error("[%s] gnutls_record_recv failed: %s (%zd)\n",
+			kr_log_verbose("[%s] gnutls_record_recv failed: %s (%zd)\n",
 				     logstring, gnutls_strerror_name(count), count);
 			return kr_error(EIO);
 		}
@@ -639,6 +639,10 @@ int tls_client_params_set(map_t *tls_client_paramlist,
 						     value, gnutls_strerror_name(res));
 					/* value will be freed at cleanup */
 					ret = kr_error(EINVAL);
+				} else {
+					kr_log_verbose("[tls_client] imported %d certs from file '%s'\n",
+							res, value);
+
 				}
 			}
 		}
@@ -790,23 +794,36 @@ skip_pins:
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
+	int ret;
+	unsigned int status;
 	for (size_t i = 0; i < ctx->params->hostnames.len; ++i) {
-		gnutls_typed_vdata_st data[2] = {
-			{ .type = GNUTLS_DT_KEY_PURPOSE_OID,
-			  .data = (void *)GNUTLS_KP_TLS_WWW_SERVER },
-			{ .type = GNUTLS_DT_DNS_HOSTNAME,
-			  .data = (void *)ctx->params->hostnames.at[i] }
-		};
-		size_t data_count = 2;
-		unsigned int status;
-		int ret = gnutls_certificate_verify_peers(ctx->c.tls_session, data,
-							  data_count, &status);
+		ret = gnutls_certificate_verify_peers3(
+				ctx->c.tls_session,
+				ctx->params->hostnames.at[i],
+				&status);
 		if ((ret == GNUTLS_E_SUCCESS) && (status == 0)) {
 			return GNUTLS_E_SUCCESS;
 		}
 	}
 
-	kr_log_error("[tls_client] failed to verify peer certificate\n");
+	if (ret == GNUTLS_E_SUCCESS) {
+		gnutls_datum_t msg;
+		ret = gnutls_certificate_verification_status_print(
+			status, gnutls_certificate_type_get(ctx->c.tls_session), &msg, 0);
+		if (ret == GNUTLS_E_SUCCESS) {
+			kr_log_error("[tls_client] failed to verify peer certificate: "
+					"%s\n", msg.data);
+			gnutls_free(msg.data);
+		} else {
+			kr_log_error("[tls_client] failed to verify peer certificate: "
+					"unable to print reason: %s (%s)\n",
+					gnutls_strerror(ret), gnutls_strerror_name(ret));
+		} /* gnutls_certificate_verification_status_print end */
+	} else {
+		kr_log_error("[tls_client] failed to verify peer certificate: "
+			     "gnutls_certificate_verify_peers3 error: %s (%s)\n",
+			     gnutls_strerror(ret), gnutls_strerror_name(ret));
+	} /* gnutls_certificate_verify_peers3 end */
 	return GNUTLS_E_CERTIFICATE_ERROR;
 }
 
@@ -873,7 +890,7 @@ int tls_client_connect_start(struct tls_client_ctx_t *client_ctx,
 	struct tls_common_ctx *ctx = &client_ctx->c;
 
 	gnutls_session_set_ptr(ctx->tls_session, client_ctx);
-	gnutls_handshake_set_timeout(ctx->tls_session, 5000);
+	gnutls_handshake_set_timeout(ctx->tls_session, KR_CONN_RTT_MAX * 3);
 	session->tls_client_ctx = client_ctx;
 	ctx->handshake_cb = handshake_cb;
 	ctx->handshake_state = TLS_HS_IN_PROGRESS;
@@ -883,7 +900,7 @@ int tls_client_connect_start(struct tls_client_ctx_t *client_ctx,
 	if (ret == GNUTLS_E_SUCCESS) {
 		return kr_ok();
 	} else if (gnutls_error_is_fatal(ret) != 0) {
-		kr_log_error("[tls_client] handshake failed (%s)\n", gnutls_strerror(ret));
+		kr_log_verbose("[tls_client] handshake failed (%s)\n", gnutls_strerror(ret));
 		return kr_error(ECONNABORTED);
 	}
 	return kr_error(EAGAIN);
