@@ -216,11 +216,13 @@ static const char * find_leq_NSEC1(struct kr_cache *cache, const struct kr_query
 	}
 	/* We know it starts before sname, so let's check the other end.
 	 * 1. construct the key for the next name - kwz_hi. */
-	const knot_dname_t *next = eh->data + 3; /* it's *full* name ATM */
+	/* it's *full* name ATM */
+	const knot_dname_t *next = eh->data + KR_CACHE_RR_COUNT_SIZE
+				 + 2 /* RDLENGTH from rfc1034 */;
 	if (!eh->data[0]) {
 		assert(false);
 		return "ERROR";
-		/* TODO: more checks?  Also, `data + 3` is kinda messy. */
+		/* TODO: more checks?  Also, `next` computation is kinda messy. */
 	}
 	/*
 	WITH_VERBOSE {
@@ -322,7 +324,7 @@ int nsec1_encloser(struct key *k, struct answer *ans,
 	knot_nsec_bitmap(&nsec_rr->rrs, &bm, &bm_size);
 
 	if (exact_match) {
-		if (kr_nsec_bitmap_nodata_check(bm, bm_size, qry->stype) != 0) {
+		if (kr_nsec_bitmap_nodata_check(bm, bm_size, qry->stype, nsec_rr->owner) != 0) {
 			assert(bm);
 			VERBOSE_MSG(qry,
 				"=> NSEC sname: match but failed type check\n");
@@ -487,7 +489,7 @@ int nsec1_src_synth(struct key *k, struct answer *ans, const knot_dname_t *clenc
 	}
 
 	/* The wildcard exists.  Find if it's NODATA - check type bitmap. */
-	if (kr_nsec_bitmap_nodata_check(bm, bm_size, qry->stype) == 0) {
+	if (kr_nsec_bitmap_nodata_check(bm, bm_size, qry->stype, nsec_rr->owner) == 0) {
 		/* NODATA proven; just need to add SOA+RRSIG later */
 		WITH_VERBOSE(qry) {
 			const char *msg_start = "=> NSEC wildcard: match proved NODATA";
