@@ -82,9 +82,16 @@ static void update_nsrep_set(struct kr_nsrep *ns, const knot_dname_t *name, uint
 
 #undef ADDR_SET
 
-static unsigned eval_addr_set(pack_t *addr_set, kr_nsrep_lru_t *rttcache, unsigned score,
-				uint8_t *addr[], struct kr_qflags opts)
+static unsigned eval_addr_set(const pack_t *addr_set, struct kr_context *ctx,
+			      unsigned score, uint8_t *addr[])
 {
+	kr_nsrep_rtt_lru_t *rtt_cache = ctx->cache_rtt;
+	struct kr_qflags opts = ctx->options;
+	kr_nsrep_rtt_lru_entry_t *rtt_cache_entry_ptr[KR_NSREP_MAXADDR] = { NULL, };
+	assert (KR_NSREP_MAXADDR >= 2);
+	unsigned rtt_cache_entry_score[KR_NSREP_MAXADDR] = { score, KR_NS_MAX_SCORE + 1, };
+	uint64_t now = kr_now();
+
 	/* Name server is better candidate if it has address record. */
 	uint8_t *it = pack_head(*addr_set);
 	while (it != pack_tail(*addr_set)) {
@@ -99,26 +106,111 @@ static unsigned eval_addr_set(pack_t *addr_set, kr_nsrep_lru_t *rttcache, unsign
 		} else {
 			is_valid = !(opts.NO_IPV4);
 		}
+
+		if (!is_valid) {
+			goto get_next_iterator;
+		}
+
 		/* Get RTT for this address (if known) */
-		if (is_valid) {
-			unsigned *cached = rttcache ? lru_get_try(rttcache, val, len) : NULL;
-			unsigned addr_score = (cached) ? *cached : KR_NS_GLUED;
-			if (addr_score < score + favour) {
-				/* Shake down previous contenders */
-				for (size_t i = KR_NSREP_MAXADDR - 1; i > 0; --i)
-					addr[i] = addr[i - 1];
-				addr[0] = it;
-				score = addr_score;
+		kr_nsrep_rtt_lru_entry_t *cached = rtt_cache ?
+						   lru_get_try(rtt_cache, val, len) :
+						   NULL;
+		unsigned cur_addr_score = KR_NS_GLUED;
+		if (cached) {
+			cur_addr_score = cached->score;
+			if (cached->score >= KR_NS_TIMEOUT) {
+				/* If NS once was marked as "timeouted",
+				 * it won't participate in NS elections
+				 * at least ctx->cache_rtt_tout_retry_interval milliseconds. */
+				uint64_t elapsed = now - cached->tout_timestamp;
+				elapsed = elapsed > UINT_MAX ? UINT_MAX : elapsed;
+				if (elapsed > ctx->cache_rtt_tout_retry_interval) {
+					/* Select this NS for probing in this particular query,
+					 * but don't change the cached score.
+					 * For other queries this NS will remain "timeouted". */
+					cur_addr_score = KR_NS_LONG - 1;
+				}
 			}
 		}
+
+		for (size_t i = 0; i < KR_NSREP_MAXADDR; ++i) {
+			if (cur_addr_score >= KR_NS_TIMEOUT) {
+				/* We can't use favour here.
+				 * If all of the conditions below are true
+				 *
+				 * rtt_cache_entry_score[i] < KR_NS_TIMEOUT
+				 * rtt_cache_entry_score[i] + favour > KR_NS_TIMEOUT
+				 * cur_addr_score < rtt_cache_entry_score[i] + favour
+				 *
+				 * we will prefer "certainly dead" cur_addr_score
+				 * instead of "almost dead, but alive" rtt_cache_entry_score[i]
+				 */
+				if (cur_addr_score >= rtt_cache_entry_score[i]) {
+					continue;
+				}
+				/* Shake down previous contenders */
+				for (size_t j = KR_NSREP_MAXADDR - 1; j > i; --j) {
+					addr[j] = addr[j - 1];
+					rtt_cache_entry_ptr[j] = rtt_cache_entry_ptr[ j - 1];
+					rtt_cache_entry_score[j] = rtt_cache_entry_score[j - 1];
+				}
+				addr[i] = it;
+				rtt_cache_entry_score[i] = cur_addr_score;
+				rtt_cache_entry_ptr[i] = cached;
+				break;
+			} else if (cur_addr_score < rtt_cache_entry_score[i] + favour) {
+				/* Shake down previous contenders */
+				for (size_t j = KR_NSREP_MAXADDR - 1; j > i; --j) {
+					addr[j] = addr[j - 1];
+					rtt_cache_entry_ptr[j] = rtt_cache_entry_ptr[j - 1];
+					rtt_cache_entry_score[j] = rtt_cache_entry_score[j - 1];
+				}
+				addr[i] = it;
+				rtt_cache_entry_score[i] = cur_addr_score;
+				rtt_cache_entry_ptr[i] = cached;
+				break;
+			}
+		}
+get_next_iterator :
 		it = pack_obj_next(it);
 	}
-	return score;
+
+	/* At this point, rtt_cache_entry_ptr contains up to KR_NSREP_MAXADDR
+	 * pointers to the rtt cache entries with the best scores for the given addr_set.
+	 * Check if there are timeouted NS. */
+
+	for (size_t i = 0; i < KR_NSREP_MAXADDR; ++i) {
+		if (rtt_cache_entry_ptr[i] == NULL)
+			continue;
+		if (rtt_cache_entry_ptr[i]->score < KR_NS_TIMEOUT)
+			continue;
+
+		uint64_t elapsed = now - rtt_cache_entry_ptr[i]->tout_timestamp;
+		elapsed = elapsed > UINT_MAX ? UINT_MAX : elapsed;
+		if (elapsed <= ctx->cache_rtt_tout_retry_interval)
+			continue;
+
+		/* rtt_cache_entry_ptr[i] points to "timeouted" rtt cache entry.
+		 * The period of the ban on participation in elections has expired. */
+
+		if (VERBOSE_STATUS) {
+			void *val = pack_obj_val(addr[i]);
+			size_t len = pack_obj_len(addr[i]);
+			char sa_str[INET6_ADDRSTRLEN];
+			int af = (len == sizeof(struct in6_addr)) ? AF_INET6 : AF_INET;
+			inet_ntop(af, val, sa_str, sizeof(sa_str));
+			kr_log_verbose("[     ][nsre] probing timeouted NS: %s, score %i\n",
+				       sa_str, rtt_cache_entry_ptr[i]->score);
+		}
+
+		rtt_cache_entry_ptr[i]->tout_timestamp = now;
+	}
+
+	return rtt_cache_entry_score[0];
 }
 
-static int eval_nsrep(const char *k, void *v, void *baton)
+static int eval_nsrep(const knot_dname_t *owner, const pack_t *addr_set, struct kr_query *qry)
 {
-	struct kr_query *qry = baton;
 	struct kr_nsrep *ns = &qry->ns;
 	struct kr_context *ctx = ns->ctx;
 	unsigned score = KR_NS_MAX_SCORE;
@@ -127,8 +219,8 @@ static int eval_nsrep(const char *k, void *v, void *baton)
 
 	/* Fetch NS reputation */
 	if (ctx->cache_rep) {
-		unsigned *cached = lru_get_try(ctx->cache_rep, k,
-					knot_dname_size((const uint8_t *)k));
+		unsigned *cached = lru_get_try(ctx->cache_rep, (const char *)owner,
+					       knot_dname_size(owner));
 		if (cached) {
 			reputation = *cached;
 		}
@@ -136,7 +228,6 @@ static int eval_nsrep(const char *k, void *v, void *baton)
 
 	/* Favour nameservers with unknown addresses to probe them,
 	 * otherwise discover the current best address for the NS. */
-	pack_t *addr_set = (pack_t *)v;
 	if (addr_set->len == 0) {
 		score = KR_NS_UNKNOWN;
 		/* If the server doesn't have IPv6, give it disadvantage. */
@@ -155,27 +246,33 @@ static int eval_nsrep(const char *k, void *v, void *baton)
 			}
 		}
 	} else {
-		score = eval_addr_set(addr_set, ctx->cache_rtt, score, addr_choice, ctx->options);
+		score = eval_addr_set(addr_set, ctx, score, addr_choice);
 	}
 
 	/* Probabilistic bee foraging strategy (naive).
 	 * The fastest NS is preferred by workers until it is depleted (timeouts or degrades),
 	 * at the same time long distance scouts probe other sources (low probability).
-	 * Servers on TIMEOUT (depleted) can be probed by the dice roll only */
-	if (score <= ns->score && (qry->flags.NO_THROTTLE || score < KR_NS_TIMEOUT)) {
-		update_nsrep_set(ns, (const knot_dname_t *)k, addr_choice, score);
+	 * Servers on TIMEOUT will not have probed at all.
+	 * Servers with score above KR_NS_LONG will have periodically removed from
+	 * reputation cache, so that kresd can reprobe them. */
+	if (score >= KR_NS_TIMEOUT) {
+		return kr_ok();
+	} else if (score <= ns->score &&
+	   (score < KR_NS_LONG  || qry->flags.NO_THROTTLE)) {
+		update_nsrep_set(ns, owner, addr_choice, score);
+		ns->reputation = reputation;
+	} else if ((kr_rand_uint(100) < 10) &&
+		   (kr_rand_uint(KR_NS_MAX_SCORE) >= score)) {
+		/* With 10% chance probe server with a probability
+		 * given by its RTT / MAX_RTT. */
+		update_nsrep_set(ns, owner, addr_choice, score);
+		ns->reputation = reputation;
+		return 1; /* Stop evaluation */
+	} else if (ns->score > KR_NS_MAX_SCORE) {
+		/* Check if any server was already selected.
+		 * If no, pick current server and continue evaluation. */
+		update_nsrep_set(ns, owner, addr_choice, score);
 		ns->reputation = reputation;
-	} else {
-		/* With 10% chance, probe server with a probability given by its RTT / MAX_RTT */
-		if ((kr_rand_uint(100) < 10) && (kr_rand_uint(KR_NS_MAX_SCORE) >= score)) {
-			/* If this is a low-reliability probe, go with TCP to get ICMP reachability check. */
-			if (score >= KR_NS_LONG) {
-				qry->flags.TCP = true;
-			}
-			update_nsrep_set(ns, (const knot_dname_t *)k, addr_choice, score);
-			ns->reputation = reputation;
-			return 1; /* Stop evaluation */
-		}
 	}
 
 	return kr_ok();
@@ -215,11 +312,11 @@ int kr_nsrep_set(struct kr_query *qry, size_t index, const struct sockaddr *sock
 
 	/* Retrieve RTT from cache */
 	struct kr_context *ctx = qry->ns.ctx;
-	unsigned *score = ctx
+	kr_nsrep_rtt_lru_entry_t *rtt_cache_entry = ctx
 		? lru_get_try(ctx->cache_rtt, kr_inaddr(sock), kr_family_len(sock->sa_family))
 		: NULL;
-	if (score) {
-		qry->ns.score = MIN(qry->ns.score, *score);
+	if (rtt_cache_entry) {
+		qry->ns.score = MIN(qry->ns.score, rtt_cache_entry->score);
 	}
 
 	return kr_ok();
@@ -240,7 +337,24 @@ int kr_nsrep_elect(struct kr_query *qry, struct kr_context *ctx)
 
 	struct kr_nsrep *ns = &qry->ns;
 	ELECT_INIT(ns, ctx);
-	return map_walk(&qry->zone_cut.nsset, eval_nsrep, qry);
+
+	int ret = kr_ok();
+	trie_it_t *it;
+	for (it = trie_it_begin(qry->zone_cut.nsset); !trie_it_finished(it);
+							trie_it_next(it)) {
+		ret = eval_nsrep(/* we trust it's a correct dname */
+				 (const knot_dname_t *)trie_it_key(it, NULL),
+				 (const pack_t *)*trie_it_val(it), qry);
+		if (ret) break;
+	}
+	trie_it_free(it);
+
+	if (qry->ns.score <= KR_NS_MAX_SCORE && qry->ns.score >= KR_NS_LONG) {
+		/* This is a low-reliability probe,
+		 * go with TCP to get ICMP reachability check. */
+		qry->flags.TCP = true;
+	}
+	return ret;
 }
 
 int kr_nsrep_elect_addr(struct kr_query *qry, struct kr_context *ctx)
@@ -252,13 +366,13 @@ int kr_nsrep_elect_addr(struct kr_query *qry, struct kr_context *ctx)
 	/* Get address list for this NS */
 	struct kr_nsrep *ns = &qry->ns;
 	ELECT_INIT(ns, ctx);
-	pack_t *addr_set = map_get(&qry->zone_cut.nsset, (const char *)ns->name);
+	pack_t *addr_set = kr_zonecut_find(&qry->zone_cut, ns->name);
 	if (!addr_set) {
 		return kr_error(ENOENT);
 	}
 	/* Evaluate addr list */
 	uint8_t *addr_choice[KR_NSREP_MAXADDR] = { NULL, };
-	unsigned score = eval_addr_set(addr_set, ctx->cache_rtt, ns->score, addr_choice, ctx->options);
+	unsigned score = eval_addr_set(addr_set, ctx, ns->score, addr_choice);
 	update_nsrep_set(ns, ns->name, addr_choice, score);
 	return kr_ok();
 }
@@ -266,14 +380,14 @@ int kr_nsrep_elect_addr(struct kr_query *qry, struct kr_context *ctx)
 #undef ELECT_INIT
 
 int kr_nsrep_update_rtt(struct kr_nsrep *ns, const struct sockaddr *addr,
-			unsigned score, kr_nsrep_lru_t *cache, int umode)
+			unsigned score, kr_nsrep_rtt_lru_t *cache, int umode)
 {
-	if (!ns || !cache || ns->addr[0].ip.sa_family == AF_UNSPEC) {
+	if (!cache || umode > KR_NS_MAX) {
 		return kr_error(EINVAL);
 	}
 
-	const char *addr_in = kr_inaddr(&ns->addr[0].ip);
-	size_t addr_len = kr_inaddr_len(&ns->addr[0].ip);
+	const char *addr_in = NULL;
+	size_t addr_len = 0;
 	if (addr) { /* Caller provided specific address */
 		if (addr->sa_family == AF_INET) {
 			addr_in = (const char *)&((struct sockaddr_in *)addr)->sin_addr;
@@ -281,31 +395,57 @@ int kr_nsrep_update_rtt(struct kr_nsrep *ns, const struct sockaddr *addr,
 		} else if (addr->sa_family == AF_INET6) {
 			addr_in = (const char *)&((struct sockaddr_in6 *)addr)->sin6_addr;
 			addr_len = sizeof(struct in6_addr);
+		} else {
+			assert(false && "kr_nsrep_update_rtt: unexpected address family");
 		}
+	} else if (ns != NULL && ns->addr[0].ip.sa_family != AF_UNSPEC) {
+		addr_in = kr_inaddr(&ns->addr[0].ip);
+		addr_len = kr_inaddr_len(&ns->addr[0].ip);
+	} else {
+		return kr_error(EINVAL);
 	}
-	unsigned *cur = lru_get_new(cache, addr_in, addr_len);
+
+	assert(addr_in != NULL && addr_len > 0);
+
+	bool is_new_entry = false;
+	kr_nsrep_rtt_lru_entry_t  *cur = lru_get_new(cache, addr_in, addr_len,
+						     (&is_new_entry));
 	if (!cur) {
 		return kr_ok();
 	}
-	/* Score limits */
-	if (score > KR_NS_MAX_SCORE) {
-		score = KR_NS_MAX_SCORE;
-	}
 	if (score <= KR_NS_GLUED) {
 		score = KR_NS_GLUED + 1;
 	}
-	/* First update is always set. */
-	if (*cur == 0) {
-		umode = KR_NS_RESET;
+	/* First update is always set unless KR_NS_UPDATE_NORESET mode used. */
+	if (is_new_entry) {
+		if (umode == KR_NS_UPDATE_NORESET) {
+			/* Zero initial value. */
+			cur->score = 0;
+		} else {
+			/* Force KR_NS_RESET otherwise. */
+			umode = KR_NS_RESET;
+		}
 	}
+	unsigned new_score = 0;
 	/* Update score, by default smooth over last two measurements. */
 	switch (umode) {
-	case KR_NS_UPDATE: *cur = (*cur + score) / 2; break;
-	case KR_NS_RESET:  *cur = score; break;
-	case KR_NS_ADD:    *cur = MIN(KR_NS_MAX_SCORE - 1, *cur + score); break;
-	case KR_NS_MAX:    *cur = MAX(*cur, score); break;
+	case KR_NS_UPDATE:
+	case KR_NS_UPDATE_NORESET:
+		new_score = (cur->score + score) / 2; break;
+	case KR_NS_RESET:  new_score = score; break;
+	case KR_NS_ADD:    new_score = MIN(KR_NS_MAX_SCORE - 1, cur->score + score); break;
+	case KR_NS_MAX:    new_score = MAX(cur->score, score); break;
 	default: break;
 	}
+	/* Score limits */
+	if (new_score > KR_NS_MAX_SCORE) {
+		new_score = KR_NS_MAX_SCORE;
+	}
+	if (new_score >= KR_NS_TIMEOUT && cur->score < KR_NS_TIMEOUT) {
+		/* Set the timestamp only when NS became "timeouted" */
+		cur->tout_timestamp = kr_now();
+	}
+	cur->score = new_score;
 	return kr_ok();
 }
 
@@ -318,7 +458,8 @@ int kr_nsrep_update_rep(struct kr_nsrep *ns, unsigned reputation, kr_nsrep_lru_t
 	/* Store in the struct */
 	ns->reputation = reputation;
 	/* Store reputation in the LRU cache */
-	unsigned *cur = lru_get_new(cache, (const char *)ns->name, knot_dname_size(ns->name));
+	unsigned *cur = lru_get_new(cache, (const char *)ns->name,
+				    knot_dname_size(ns->name), NULL);
 	if (cur) {
 		*cur = reputation;
 	}
@@ -339,9 +480,9 @@ int kr_nsrep_copy_set(struct kr_nsrep *dst, const struct kr_nsrep *src)
 	return kr_ok();
 }
 
-int kr_nsrep_sort(struct kr_nsrep *ns, kr_nsrep_lru_t *cache)
+int kr_nsrep_sort(struct kr_nsrep *ns, kr_nsrep_rtt_lru_t *rtt_cache)
 {
-	if (!ns || !cache) {
+	if (!ns || !rtt_cache) {
 		assert(false);
 		return kr_error(EINVAL);
 	}
@@ -364,22 +505,24 @@ int kr_nsrep_sort(struct kr_nsrep *ns, kr_nsrep_lru_t *cache)
 		if (sa->sa_family == AF_UNSPEC) {
 			break;
 		}
-		unsigned *score = lru_get_try(cache, kr_inaddr(sa),
-						kr_family_len(sa->sa_family));
-		if (!score) {
+		kr_nsrep_rtt_lru_entry_t *rtt_cache_entry = lru_get_try(rtt_cache,
+									kr_inaddr(sa),
+									kr_family_len(sa->sa_family));
+		if (!rtt_cache_entry) {
 			scores[i] = 1; /* prefer unknown to probe RTT */
-		} else if ((kr_rand_uint(100) < 10)
-				&& (kr_rand_uint(KR_NS_MAX_SCORE) >= *score)) {
+		} else if ((kr_rand_uint(100) < 10) &&
+			   (kr_rand_uint(KR_NS_MAX_SCORE) >= rtt_cache_entry->score)) {
 			/* some probability to bump bad ones up for re-probe */
 			scores[i] = 1;
 		} else {
-			scores[i] = *score;
+			scores[i] = rtt_cache_entry->score;
 		}
 		if (VERBOSE_STATUS) {
 			char sa_str[INET6_ADDRSTRLEN];
 			inet_ntop(sa->sa_family, kr_inaddr(sa), sa_str, sizeof(sa_str));
 			kr_log_verbose("[     ][nsre] score %d for %s;\t cached RTT: %d\n",
-					scores[i], sa_str, score ? *score : -1);
+					scores[i], sa_str,
+					rtt_cache_entry ? rtt_cache_entry->score : -1);
 		}
 	}
 
