@@ -53,7 +53,7 @@ struct kr_qflags {
 	bool BADCOOKIE_AGAIN : 1;/**< Query again because bad cookie returned. */
 	bool CNAME : 1;          /**< Query response contains CNAME in answer section. */
 	bool REORDER_RR : 1;     /**< Reorder cached RRs. */
-	bool TRACE : 1;          /**< Log answer with kr_verbose_log(), unless -DNDEBUG. */
+	bool TRACE : 1;          /**< Also log answers if --verbose. */
 	bool NO_0X20 : 1;        /**< Disable query case randomization . */
 	bool DNSSEC_NODS : 1;    /**< DS non-existance is proven */
 	bool DNSSEC_OPTOUT : 1;  /**< Closest encloser proof has optout */
@@ -62,6 +62,7 @@ struct kr_qflags {
 	bool FORWARD : 1;        /**< Forward all queries to upstream; validate answers. */
 	bool DNS64_MARK : 1;     /**< Internal mark for dns64 module. */
 	bool CACHE_TRIED : 1;    /**< Internal to cache module. */
+	bool NO_NS_FOUND : 1;    /**< No valid NS found during last PRODUCE stage. */
 };
 
 /** Combine flags together.  This means set union for simple flags. */
